@@ -16,6 +16,8 @@ struct MVMNativeCallBody {
     MVMint16   *arg_types;
     MVMObject **arg_info;
     MVMJitCode *jitcode;
+    MVMObject  *resolve_lib_name;
+    MVMObject  *resolve_lib_name_arg;
 };
 
 struct MVMNativeCall {
