@@ -11,9 +11,9 @@
  * What we need is to round the value rounded up to the next power of 2, and
  * then the log base 2 of that. Don't call this with v == 0. */
 MVMuint32 MVM_round_up_log_base2(MVMuint32 v) {
-    static const unsigned int MultiplyDeBruijnBitPosition[32] = {
-        0, 9, 1, 10, 13, 21, 2, 29, 11, 14, 16, 18, 22, 25, 3, 30,
-        8, 12, 20, 28, 15, 17, 24, 7, 19, 27, 23, 6, 26, 5, 4, 31
+    static const uint8_t MultiplyDeBruijnBitPosition[32] = {
+        1, 10, 2, 11, 14, 22, 3, 30, 12, 15, 17, 19, 23, 26, 4, 31,
+        9, 13, 21, 29, 16, 18, 25, 8, 20, 28, 24, 7, 27, 6, 5, 32
     };
 
     /* this rounds (v - 1) down to one less than a power of 2 */
@@ -24,7 +24,7 @@ MVMuint32 MVM_round_up_log_base2(MVMuint32 v) {
     v |= v >> 8;
     v |= v >> 16;
 
-    return 1 + MultiplyDeBruijnBitPosition[(MVMuint32)(v * 0x07C4ACDDU) >> 27];
+    return MultiplyDeBruijnBitPosition[(uint32_t)(v * 0x07C4ACDDU) >> 27];
 }
 
 MVM_STATIC_INLINE MVMuint32 hash_true_size(MVMStrHashTable *hashtable) {
@@ -136,15 +136,6 @@ MVM_STATIC_INLINE struct MVMStrHashHandle *hash_insert_internal(MVMThreadContext
 
             if (*metadata == 0) {
                 /* Open goal. Score! */
-                /* However, we can still exceed the maximum probe distance.
-                 * Optimisation from Martin Ankerl's implementation:
-                 * setting this to zero forces a resize on any insert, *before*
-                 * the actual insert, so that we never end up having to handle
-                 * overflow *during* this loop. This loop can always complete.
-                 */
-                if (probe_distance == MVM_HASH_MAX_PROBE_DISTANCE) {
-                    hashtable->max_items = 0;
-                }
             } else {
                 /* make room. */
 
@@ -160,8 +151,11 @@ MVM_STATIC_INLINE struct MVMStrHashHandle *hash_insert_internal(MVMThreadContext
                 do {
                     MVMuint8 new_probe_distance = 1 + old_probe_distance;
                     if (new_probe_distance == MVM_HASH_MAX_PROBE_DISTANCE) {
-                        /* Again, without action, any insertion might cause us
-                         * to excede our probe distance. */
+                        /* Optimisation from Martin Ankerl's implementation:
+                           setting this to zero forces a resize on any insert,
+                           *before* the actual insert, so that we never end up
+                           having to handle overflow *during* this loop. This
+                           loop can always complete. */
                         hashtable->max_items = 0;
                     }
                     /* a swap: */
@@ -174,6 +168,19 @@ MVM_STATIC_INLINE struct MVMStrHashHandle *hash_insert_internal(MVMThreadContext
                         hashtable->entry_size * entries_to_move);
             }
 
+            /* The same test and optimisation as in the "make room" loop - we're
+             * about to insert something at the (current) max_probe_distance, so
+             * signal to the next insertion that it needs to take action first.
+             */
+            if (probe_distance == MVM_HASH_MAX_PROBE_DISTANCE) {
+                hashtable->max_items = 0;
+            }
+
+#if HASH_DEBUG_ITER
+            ++hashtable->serial;
+            hashtable->last_delete_at = 0;
+#endif
+
             *metadata = probe_distance;
             struct MVMStrHashHandle *entry = (struct MVMStrHashHandle *) entry_raw;
             entry->key = NULL;
@@ -317,6 +324,14 @@ void MVM_str_hash_delete_nocheck(MVMThreadContext *tc,
                 *metadata_target = 0;
                 --hashtable->cur_items;
 
+#if HASH_DEBUG_ITER
+            ++hashtable->serial;
+            /* We need to calculate the interator position corresponding to the
+             * hash bucket we actually used. `metadata - hashtable->metadata
+             * gives the bucket, and iterators store `$bucket + 1`, hence: */
+            hashtable->last_delete_at = 1 + metadata - hashtable->metadata;
+#endif
+
                 /* Job's a good 'un. */
                 return;
             }
