@@ -339,11 +339,11 @@ static MVMuint8 windows1251_cp_to_char(MVMint32 codepoint) {
 /* Decodes using a decodestream. Decodes as far as it can with the input
  * buffers, or until a stopper is reached. */
 MVMuint32 MVM_string_windows125X_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
-                                         const MVMint32 *stopper_chars,
+                                         const MVMuint32 *stopper_chars,
                                          MVMDecodeStreamSeparators *seps,
                                          const MVMuint16 *codetable) {
-    MVMint32 count = 0, total = 0;
-    MVMint32 bufsize;
+    MVMuint32 count = 0, total = 0;
+    MVMuint32 bufsize;
     MVMGrapheme32 *buffer = NULL;
     MVMDecodeStreamBytes *cur_bytes = NULL;
     MVMDecodeStreamBytes *last_accept_bytes = ds->bytes_head;
@@ -371,7 +371,7 @@ MVMuint32 MVM_string_windows125X_decodestream(MVMThreadContext *tc, MVMDecodeStr
     while (cur_bytes) {
         /* Process this buffer. */
         MVMint32  pos = cur_bytes == ds->bytes_head ? ds->bytes_head_pos : 0;
-        unsigned char *bytes = (unsigned char *)cur_bytes->bytes;
+        MVMuint8 *bytes = cur_bytes->bytes;
         while (pos < cur_bytes->length || repl_pos) {
             MVMGrapheme32 graph;
             MVMCodepoint codepoint = codetable[bytes[pos++]];
@@ -462,14 +462,14 @@ MVMuint32 MVM_string_windows125X_decodestream(MVMThreadContext *tc, MVMDecodeStr
 /* Decodes using a decodestream. Decodes as far as it can with the input
  * buffers, or until a stopper is reached. */
 MVMuint32 MVM_string_windows1252_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
-                                         const MVMint32 *stopper_chars,
+                                         const MVMuint32 *stopper_chars,
                                          MVMDecodeStreamSeparators *seps) {
     return MVM_string_windows125X_decodestream(tc, ds, stopper_chars, seps, windows1252_codepoints);
 }
 /* Decodes using a decodestream. Decodes as far as it can with the input
  * buffers, or until a stopper is reached. */
 MVMuint32 MVM_string_windows1251_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
-                                         const MVMint32 *stopper_chars,
+                                         const MVMuint32 *stopper_chars,
                                          MVMDecodeStreamSeparators *seps) {
     return MVM_string_windows125X_decodestream(tc, ds, stopper_chars, seps, windows1251_codepoints);
 }
@@ -481,12 +481,11 @@ MVMString * MVM_string_windows125X_decode(MVMThreadContext *tc,
         const MVMObject *result_type, char *windows125X_c, size_t bytes,
         MVMString *replacement, const MVMuint16 *codetable, MVMint64 config) {
     MVMuint8 *windows125X = (MVMuint8 *)windows125X_c;
-    MVMString *result = (MVMString *)REPR(result_type)->allocate(tc, STABLE(result_type));
+    MVMString *result;
     size_t pos, result_graphs, additional_bytes = 0;
     MVMStringIndex repl_length = replacement ? MVM_string_graphs(tc, replacement) : 0;
 
-    result->body.storage_type    = MVM_STRING_GRAPHEME_32;
-    result->body.storage.blob_32 = MVM_malloc(sizeof(MVMGrapheme32) * bytes);
+    MVMGrapheme32 *buffer = MVM_malloc(sizeof(MVMGrapheme32) * bytes);
 
     result_graphs = 0;
     for (pos = 0; pos < bytes; pos++) {
@@ -501,16 +500,15 @@ MVMString * MVM_string_windows125X_decode(MVMThreadContext *tc,
                 /* Since things we are decoding always fit into Unicode, if we are
                  * using a replacement, it won't get used unless we use strict */
                 if (replacement && MVM_ENCODING_CONFIG_STRICT(config)) {
-                    int i = 0;
+                    MVMStringIndex i = 0;
                     /* Only triggered if repl_length > 1. Copies all but the last
                      * grapheme in the replacement string */
                     if (1 < repl_length) {
                         additional_bytes += repl_length - 1;
-                        result->body.storage.blob_32 = realloc(result->body.storage.blob_32,
-                            sizeof(MVMGrapheme32) * (additional_bytes + bytes));
+                        buffer = realloc(buffer, sizeof(MVMGrapheme32) * (additional_bytes + bytes));
                         for (; i < repl_length - 1; i++) {
                             MVMGrapheme32 graph = MVM_string_get_grapheme_at(tc, replacement, i);
-                            result->body.storage.blob_32[result_graphs++] = graph;
+                            buffer[result_graphs++] = graph;
                         }
                     }
                     /* Now we set `codepoint` to the last grapheme in the replacement
@@ -518,6 +516,7 @@ MVMString * MVM_string_windows125X_decode(MVMThreadContext *tc,
                     codepoint = MVM_string_get_grapheme_at(tc, replacement, i);
                 }
                 else if (MVM_ENCODING_CONFIG_STRICT(config)) {
+                    MVM_free(buffer);
                     /* Throw an exception if that codepoint has no mapping */
                     char *enc_name = codetable == windows1252_codepoints
                         ? "Windows-1252" : "Windows-1251";
@@ -534,8 +533,11 @@ MVMString * MVM_string_windows125X_decode(MVMThreadContext *tc,
                 }
             }
         }
-        result->body.storage.blob_32[result_graphs++] = codepoint;
+        buffer[result_graphs++] = codepoint;
     }
+    result = (MVMString *)REPR(result_type)->allocate(tc, STABLE(result_type));
+    result->body.storage.blob_32 = buffer;
+    result->body.storage_type = MVM_STRING_GRAPHEME_32;
     result->body.num_graphs = result_graphs;
 
     return result;
